#include "debug.hpp"

void export_lattice(std::ostream& os, const Lattice& lat) {
    os << "#BEGIN LATTICE" << '\n';
    for (int row = 0; row < lat.data_height; row++) {
        for (int col = 0; col < lat.data_width; col++) {
            os << std::fixed << std::setprecision(0) << std::setw(6) << row << " " << std::setw(6) << col << " ";
            os << lat.atd(row, col) << '\n';
        }
    } 
    os << "#END LATTICE" << '\n';
}

void export_sections(std::ostream& os, const std::vector<Section>& sec) {
    os << "#BEGIN SECTIONS" << '\n';
    for (int i = 0; i < sec.size(); i++) {
        os << std::setw(6) << i << " " << sec[i] << '\n';
    }
    os << "#END SECTIONS" << '\n';
}

void export_section_exchange(std::ostream& os, const std::vector<Section_exchange>& sec) {
    os << "#BEGIN EXCHANGE" << '\n';
    for (int i = 0; i < sec.size(); i++) {
        os << std::setw(6) << i << "\n" << "SEND" << "\n";
        for(auto ex: sec[i].send) {
            os << std::setw(6) << ex.where << " " << std::setw(6) << ex.tag << " " << ex.section << '\n';
        }
        os << "RECV" << '\n';
        for(auto ex: sec[i].recv) {
            os << std::setw(6) << ex.where << " " << std::setw(6) << ex.tag << " " << ex.section << '\n';
        }
        os << '\n';
    }
    os << "#END EXCHANGE" << '\n';
}

void export_rank_assign(std::ostream& os, const std::vector<std::vector<int>>& assign) {
    os << "#BEGIN RANKASSIGN" << '\n';
    for (int i = 0; i < assign.size(); i++) {
        os << i << " ";
        for(int i: assign[i]) {
            os << std::setw(6) << i << " ";
        }
        os << '\n';
    }
    os << "#END RANKASSIGN" << '\n';
}

void export_global_state(   const Lattice& lattice, 
                            const std::vector<Section>& sections, 
                            const std::vector<Section_exchange>& all_exchanges, 
                            const std::vector<std::vector<int>>& rank_assignment) {
    // Write complete lattice state;
    std::string filename = "global_export.txt";
    std::ofstream export_ofs{filename};

    if(!export_ofs) { std::cerr << "Error opening file"; return; }

    export_lattice(export_ofs, lattice);
    export_sections(export_ofs, sections);
    export_section_exchange(export_ofs, all_exchanges);
    export_rank_assign(export_ofs, rank_assignment);
}

void export_sections_as_tikz(
                            const Lattice& lattice, 
                            const std::vector<Section>& sections, 
                            const std::vector<Section_exchange>& all_exchanges, 
                            const std::vector<std::vector<int>>& rank_assignment,
                            const std::string filename) 
{
    // Write complete lattice state;
    std::ofstream export_ofs{filename};

    std::vector<std::string> colors {"black", "blue", "red", "green", "orange"};
    export_ofs << "\\documentclass{article}\n\\usepackage{tikz}\n\\begin{document}\n\\begin{tikzpicture}\n";
    export_ofs << "\\draw[help lines] (0, 0) grid (" << lattice.width << "," << lattice.height << ");\n";
    // We'll loop though this from the rank assignment point of view to keep track of colors
    for (int i = 0; i < rank_assignment.size(); i++) {
        for (int a: rank_assignment[i]) {
            const auto& sec =  sections[a];

            // We'll write out a section
            char section_rect_str[100];
            sprintf(section_rect_str, "\\draw[%s, thick] (%d, %d) rectangle (%d, %d);", colors[i % colors.size()].c_str(), sec.tl_col, sec.tl_row, sec.br_col + 1, sec.br_row + 1);
            export_ofs << section_rect_str << "\n";

            // Map exchanges for this section
            auto& sec_exch  = all_exchanges[a];
            for (auto exch: sec_exch.send) {
                const auto& exch_sec = exch.section;

                char section_rect_str[100];
                sprintf(section_rect_str, "\\draw[%s, thick] (%d, %d) rectangle (%d, %d);", colors[(i+1) % colors.size()].c_str(), sec.tgc(exch_sec.tl_col), sec.tgr(exch_sec.tl_row), sec.tgc(exch_sec.br_col + 1), sec.tgr(exch_sec.br_row + 1));
                export_ofs << section_rect_str << "\n";
            }
        }
    }

    export_ofs << "\\end{tikzpicture}\n\\end{document}\n";
}