#include "exch_plan.hpp"


int to_global_row (const Lattice &l, const Section &s, int row, bool periodic = false) 
{
    int g_row = s.to_global_row(row);
    if (periodic) {
        return modulo(g_row, l.height);
    }
    else {
        return g_row;
    }
}

int to_global_col (const Lattice &l, const Section &s, int col, bool periodic = false) 
{
    int g_col = s.to_global_col(col);
    if (periodic) {
        return modulo(g_col, l.height);
    }
    else {
        return g_col;
    }
}

void tag_section_id (std::vector<std::vector<int>>& map, std::vector<Section>& sections)
{
    for (int i = 0; i < sections.size(); i++)
    {
        const auto& s = sections[i];
        for (int row = 0; row < s.height; row++)
        {
            for (int col = 0; col < s.width; col++) {
                map[s.to_global_row(row)][s.to_global_col(col)] = i;
            }
        }
    }
}

bool is_new_sec(int old_val, int new_val)
{
    return old_val != new_val;
}

bool should_add_sec(int sec_val) 
{
    return sec_val != -1;
}

std::vector<std::vector<std::vector<int>>> get_borders(int height, int width) {
    std::vector<std::vector<std::vector<int>>> borders {
            // Horiz
            {{-1, 0},       {-1, width - 1}, {0}},           // North
            {{height, 0}, {height, width-1}, {0}},     // South
            // Vert
            {{0, -1},      {height - 1, -1}, {1}},            // West
            {{0, width}, {height-1,width}, {1}},        // East
            // Corn left top, bot, right top bot
            {{-1, -1}, {-1, -1}, {0}},
            {{height, -1}, {height, -1}, {0}},
            {{-1, width}, {-1, width}, {0}},
            {{height,width}, {height,width}, {0}}
        };
    return borders;
}

void add_exchange(Lattice& lattice, std::vector<Section_exchange>& exchanges, std::vector<Section>& sections, int recv_sec_id, int send_sec_id, int tl_row, int tl_col, int br_row, int br_col, int tag)
{
    // Add section determined by (cur_row_start, cur_col_start), (previous_row, previous_col)
   
    // Receive first
    exchanges[recv_sec_id].recv.push_back(Exchange({{tl_row, tl_col}, {br_row, br_col}}, send_sec_id, tag));

    // Now we need to convert coords into global, and then back to local
    Section& recv_sec = sections[recv_sec_id];
    int g_tl_row = to_global_row(lattice, recv_sec, tl_row, true);
    int g_tl_col = to_global_col(lattice, recv_sec, tl_col, true);
    int g_br_row = to_global_row(lattice, recv_sec, br_row, true);
    int g_br_col = to_global_col(lattice, recv_sec, br_col, true);

    Section& send_sec = sections[send_sec_id];
    int s_tl_row = send_sec.from_global_row(g_tl_row);
    int s_tl_col = send_sec.from_global_col(g_tl_col);
    int s_br_row = send_sec.from_global_row(g_br_row);
    int s_br_col = send_sec.from_global_col(g_br_col);

    exchanges[send_sec_id].send.push_back(Exchange({{s_tl_row, s_tl_col}, {s_br_row, s_br_col}}, recv_sec_id, tag));
}

std::vector<Section_exchange> determine_exchanges(std::vector<Section>& sections, Lattice& lattice) {

    auto section_map = std::vector<std::vector<int>>(lattice.height, std::vector<int>(lattice.width, -1));
    
    // For each section, we iterate over all of it and tag the index of this section
    tag_section_id(section_map, sections);

    // Generate empty containers for the exchanges
    std::vector<Section_exchange> exchanges(sections.size());

    // Initialize tag to be 1000
    int tag = 1000;

    // Determine exchanges for each section
    for (int i = 0; i < sections.size(); i++)
    {
        using namespace std;
        const auto& s = sections[i];
        
        auto borders = get_borders(s.height, s.width);

        for (const auto& border_coords: borders) 
        {
            
            // Current section
            int cur_id = -1;
            int cur_row_start = -1;
            int cur_col_start = -1;

            int previous_row = -1;
            int previous_col = -1;

            for (int row = border_coords[0][0]; row <= border_coords[1][0]; row++) 
            {
                for (int col = border_coords[0][1]; col <= border_coords[1][1]; col++)
                {
                    // Determine the global coords of this border pixel
                    int g_row = to_global_row(lattice, s, row, true);
                    int g_col = to_global_col(lattice, s, col, true);

                    // Get the section we're currently looking at
                    int sec = section_map[g_row][g_col];

                    // Check if this is a change
                    if (sec != cur_id) {

                        // If previous was positive, it was a section that should be exchanged
                        if (cur_id >= 0) {

                            // Add section determined by (cur_row_start, cur_col_start), (previous_row, previous_col)
                            add_exchange(lattice, exchanges, sections, i, cur_id, cur_row_start, cur_col_start, previous_row, previous_col, tag);
                            tag++;
                        }
                        
                        cur_id = sec;
                        cur_row_start = row;
                        cur_col_start = col;
                    }

                    // Finally, update the previous to be current
                    previous_col = col;
                }

                previous_row = row;
            }

            // We may exit here with a cur_id != -1, which means we still need to add that exchange
            // Add section determined by (cur_row_start, cur_col_start), (previous_row, previous_col)
            if (cur_id != -1) {
                add_exchange(lattice, exchanges, sections, i, cur_id, cur_row_start, cur_col_start, previous_row, previous_col, tag);
                tag++;
            }
        }
    }

    return exchanges;    
}




/*******************************************************************
 * 
 * 
 *  Testing
 * 
 * 
 * ****************************************************************/

// Equality ignores tag
bool operator==(const Exchange& lhs, const Exchange& rhs) {
    return  lhs.where == rhs.where && 
            lhs.section == rhs.section;
}

// Determine if all exchanges in first param is in the second param.
bool section_exchange_subset(const Section_exchange& lhs, const Section_exchange& rhs) {
    // Compare sends
    for (auto exch: lhs.send) {
        auto match = std::find(rhs.send.begin(), rhs.send.end(), exch);
        if (match == rhs.send.end()) {
            return false;
        }
    }

    for (auto exch: lhs.recv) {
        auto match = std::find(rhs.recv.begin(), rhs.recv.end(), exch);
        if (match == rhs.recv.end()) {
            return false;
        }
    }

    return true;
}

bool section_exchange_equal(const Section_exchange& lhs, const Section_exchange &rhs) {
    return section_exchange_subset(lhs, rhs) && section_exchange_subset(rhs, lhs);
}

bool operator==(const Section_exchange& lhs, const Section_exchange &rhs) {
    return section_exchange_equal(lhs, rhs);
}

/**
 * Above the test cases, a small illustration attempts to show
 * what scenario is tested. O is fluid, x is solid.
 * 
 * */


std::ostream& operator<<(std::ostream& os, const Section_exchange& sec) {
    os << "Send:\n";
    for(auto exch: sec.send) {
        os << "[ " << exch.section << " ]" << '\n';
    }
    os << "Recv:\n";
    for(auto exch: sec.recv) {
        os << "[ " << exch.section << " ]" << '\n';
    }
    return os;
}


TEST_CASE("Get borders are correct") 
{
    auto borders = get_borders(4, 4);
    CHECK_EQ(borders[0][0][0], -1);
    CHECK_EQ(borders[0][0][1], 0);
}

TEST_CASE("Correct exchanges single section")
{
    Lattice l{4,4};
    std::vector<Section> sections{ 
        Section{0, 0, 4, 4}
    };

    auto plan = determine_exchanges(sections, l);

    // Expected exchanges
    // We should have 8 sends and 8 recvs. 
    // One per side plus one per corner
    Section_exchange zero{
        // Send
        {
            // Horiz
            {{{0,0}, {0,3}}, 0, 0},
            {{{3,0}, {3,3}}, 0, 0},
            // Vert
            {{{0,0}, {3,0}}, 0, 0},
            {{{0,3}, {3,3}}, 0, 0},
            // Corners
            {{{0,0}, {0,0}}, 0, 0},
            {{{0,3}, {0,3}}, 0, 0},
            {{{3,0}, {3,0}}, 0, 0},
            {{{3,3}, {3,3}}, 0, 0},
        },

        //Recv
        {
            // Horiz
            {{{4,0}, {4,3}}, 0, 0},
            {{{-1,0}, {-1,3}}, 0, 0},
            // Vert
            {{{0,4}, {3,4}}, 0, 0},
            {{{0,-1}, {3,-1}}, 0, 0},
            // Corners
            {{{-1,-1}, {-1,-1}}, 0, 0},
            {{{-1,4}, {-1,4}}, 0, 0},
            {{{4,-1}, {4,-1}}, 0, 0},
            {{{4,4}, {4,4}}, 0, 0},
        },
    };

    CHECK_EQ(plan[0], zero);
}

// /**
//  * ---------
//  * |O O O O|
//  * |O O O O|
//  * ---------
//  * |O O O O|
//  * |O O O O|
//  * ---------
//  * */
// // Horizontal split across the entire lattice
// TEST_CASE("PlanExchangeCorrectHorizontalBorder") {
//     Lattice l{4,4};
//     std::vector<Section> sections{ 
//         Section{0, 0, 2, 4},
//         Section{2, 0, 2, 4}
//     };

//     auto plan = determine_exchanges(sections, l);
    
//     // Expected exchanges
//     // Section 0, upper section
//     Section_exchange zero{
//         {{{{1,0}, {1,3}}, 1, 0}},
//         {{{{2,0}, {2,3}}, 1, 0}}
//     };

//     // Section 1, lower section
//     Section_exchange one {
//         {{{{0,0}, {0,3}}, 0, 0}},
//         {{{{-1,0}, {-1, 3}}, 0, 0}}
//     };

//     CHECK(section_exchange_equal(plan[0], zero));
//     CHECK(section_exchange_equal(plan[1], one));
// }

// /**
//  * -----|-----
//  * |O O | O O|
//  * |O O | O O|
//  * |O O | O O|
//  * |O O | O O|
//  * -----|-----
//  * */
// // Vertical split across the entire lattice
// TEST_CASE("PlanExchangeCorrectVerticalBorder") {
//     Lattice l{4,4};
//     std::vector<Section> sections{ 
//         Section{0, 0, 4, 2},
//         Section{0, 2, 4, 2}
//     };

//     // Expected exchanges
//     // Section 0, upper section
//     Section_exchange zero{
//         {{{{0,1}, {3,1}}, 1, 0}},
//         {{{{0,2}, {3,2}}, 1, 0}}
//     };

//     // Section 1, lower section
//     Section_exchange one {
//         {{{{0,0}, {3,0}}, 0, 0}},
//         {{{{0,-1}, {3, -1}}, 0, 0}}
//     };


//     auto plan = determine_exchanges(sections, l);

//     CHECK(plan[0] == zero);
//     CHECK(plan[1] == one);
// }

// // Vertical split across part of the lattice
// TEST_CASE("PlanExchangeCorrectShortVerticalBorder") {
//     Lattice l{6,6};
//     std::vector<Section> sections{ 
//         Section{1, 1, 4, 2},
//         Section{1, 3, 4, 2}
//     };

//     // Expected exchanges
//     // Section 0, upper section
//     Section_exchange zero{
//         {{{{0,1}, {3,1}}, 1, 0}},
//         {{{{0,2}, {3,2}}, 1, 0}}
//     };

//     // Section 1, lower section
//     Section_exchange one {
//         {{{{0,0}, {3,0}}, 0, 0}},
//         {{{{0,-1}, {3, -1}}, 0, 0}}
//     };

//     std::vector<Section_exchange> sec_exch = {zero, one};
//     std::vector<std::vector<int>> rank_assign = {{0}, {1}};


//     auto plan = determine_exchanges(sections, l);
//     //export_sections_as_tikz(l, sections, plan, rank_assign, "shortvert.tex");

//     CHECK(plan[0] == zero);
//     CHECK(plan[1] == one);
// }