#pragma once

#include <iostream>
#include <iomanip>
#include <memory>
#include <cmath>

#include "indexing.hpp"

enum class Point_type
{
    Fluid,
    Solid
};

/* Lattice point data structure */
struct Lpoint
{
    Point_type type = Point_type::Fluid;

    float velocity[2] = {0, 0};
    /* Density vectors */
    float density[6][2] = {{0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}, {0, 0}};
};


std::ostream& operator<<(std::ostream& os, const Lpoint& lp);


/**
 * @brief A lattice with a border around. 
 * 
 * 
 */
struct Lattice
{
    // Underlying storage, including border for exchange
    Lpoint *data; // Use unique_ptr to avoid own impl of delete and move
    int data_height, data_width;

    // Lattice size
    int height, width;

    Lattice(): height(0), width(0), data_height(0), data_width(0), data(nullptr) { }

    // Create an empty lattice, no init
    Lattice(int h, int w): height(h), width(w),
                                    data_height(h + 2), data_width(w + 2)
    {
        data = new Lpoint[data_height * data_width];
    };

    // Delete
    ~Lattice()
    {
        delete[] data;
    }

    // Remove copy assignment and constructor. Moves will have to be explicit, 
    // and if needed, a copy function can be made to ensure no copies are unintended.
    Lattice(const Lattice& other) = delete;
    Lattice& operator=(const Lattice&) = delete; 


    // Move const
    Lattice(Lattice &&other) noexcept : height(other.height), width(other.width),
                                        data_height(other.height + 2), data_width(other.width + 2)
    {
        data = other.data;
        other.data = nullptr;
    }

    // Move assign
    Lattice& operator=(Lattice &&other) noexcept
    {
        std::swap(data, other.data);
        std::swap(data_height, other.data_height);
        std::swap(data_width, other.data_width);
        std::swap(height, other.height);
        std::swap(width, other.width);
        return *this;
    }

    Lpoint& operator[](int i) { return data[i]; }

    Lpoint& at(int row, int col) { return data[(row + 1) * (data_width) + (col + 1)]; }
    Lpoint& atd(int row, int col) { return data[row * data_width + col]; }

    const Lpoint& at(int row, int col) const { return data[(row + 1) * (data_width) + (col + 1)]; }
    const Lpoint& atd(int row, int col) const { return data[row * data_width + col]; }

    Lpoint* atp(int row, int col) { return &(data[row * data_width + col]); }
};


void initialize_even_distribution(Lattice &l);

std::ostream& write_lattice_to_stream(std::ostream& os, const Lattice& lat);
