#include <unistd.h>
#include <iostream>
#include <fstream>
#include <sstream>
#include <chrono>

#include <mpi.h>

// Assert from doctest
#define DOCTEST_CONFIG_IMPLEMENT
#include "doctest.hpp"

#include "CLI11.hpp"

#include "raster.hpp"
#include "lbm_kernel.hpp"

#include "part_grid.hpp"
#include "part_rec.hpp"
#include "part_rec2.hpp"

#include "scheduling.hpp"
#include "exch_plan.hpp"

#include "mpi_types.hpp"
#include "mpi_exchange.hpp"

#include "pbm.h"
#include "debug.hpp"

#ifdef _OPENMP 
#include "omp.h"
#endif // OPENMP

class Channel_force {
    int lower_limit;
    int upper_limit;

    const float ext_force[2] {0.0, 5e-5};
public:
    Channel_force(int lower, int upper): lower_limit(lower), upper_limit(upper) {}
    void operator()(Section& section, int row, int col, int i, float& delta_N) {
        if ( section.to_global_row(row) >= lower_limit && section.to_global_row(row) < upper_limit ) {
            delta_N += (1.0/3.0) * (c[i][0]*ext_force[0] + c[i][1]*ext_force[1]);
        }
    }
};

class No_force {
public:
    void operator()(Section& section, int row, int col, int i, float& delta_N) {}
};

enum class External_force { NONE, CHANNEL };

int main(int argc, char **argv) {
    // construct a context
    doctest::Context context;
    context.applyCommandLine(argc, argv);
    
    // sets the context as the default one - so asserts used outside of a testing context do not crash
    context.setAsDefaultForAssertsOutOfTestCases();

    int res = context.run(); // run

    if(context.shouldExit()) // important - query flags (and --exit) rely on the user doing this
        return res;          // propagate the result of the tests
    
    int client_stuff_return_code = 0;

    // Start client code
    int rank, size;

    MPI_Init(&argc, &argv);
    // We create an extra scope to ensure everything will be destructed before the finalize call
    {
        MPI_Comm_rank(MPI_COMM_WORLD, &rank);
        MPI_Comm_size(MPI_COMM_WORLD, &size);

        // Initialize types
        init_mpi_types();
        // A MPI type factory for all the different lattice sizes
        Lattice_type_factory lat_mpi_type_fac(lpoint_type);

#pragma region // Local variables, sections, lattices, etc
        // Only valid on rank 0
        Lattice global_lattice;
        Section global_section;
        std::vector<Section> all_sections;
        std::vector<Section_exchange> all_exchanges;
        std::vector<std::vector<int>> rank_assignment; // This is vector[rank] -> vector[section]. 
        
        // Sent from master to all slaves
        int section_count;
        std::vector<int> section_ids;
        std::vector<Section> sections;
        std::vector<Section_exchange> exchanges;
        std::vector<Lattice> lattices;

        // Scattered from master
        std::vector<int> section_assignment; // This is vector[section] -> rank 
#pragma endregion // Local variables, sections, lattices, etc

#pragma region // CLI arguments and parser
        int MAX_ITER = 100;
        int channel_force_lower = 1;
        int channel_force_upper = 100;
        std::string force_type = "none";
        External_force external_force = External_force::NONE;
        std::string lattice_file = "samples/moffat400x400.pbm";

        // Parser
        CLI::App app{"D2Q6 Lattice Boltzmann proxy application"};
        
        app.add_option("-i, --iter", MAX_ITER, "Number of iterations");

        // CLI external force method and arguments
        // Channel force params
        auto cf_lower_opt = app.add_option("--cf-lower", channel_force_lower, "Lower limit of the channel force, if used");
        auto cf_upper_opt = app.add_option("--cf-upper", channel_force_upper, "Upper limit of the channel force, if used");

        auto cf_opt = app.add_option("--cf", force_type, "Use channel force, default is none, accepted is channel.");

        // Sectioning

        int part_grid_width = 200;
        int part_grid_height = 200; 
        bool use_grid_part = false;

        auto part_grid_width_opt = app.add_option("--part-grid-width", part_grid_width, "Maximum grid section width");
        auto part_grid_height_opt = app.add_option("--part-grid-height", part_grid_height, "Maximum grid section height");
        auto use_grid_part_opt = app.add_flag("--part-grid", use_grid_part, "Use grid partitioning");

        int part_rec_width = 200;
        int part_rec_height = 200;
        bool use_rec_part = false;

        auto part_rec_width_opt = app.add_option("--part-rec-width", part_rec_width, "Maximum rec section width");
        auto part_rec_height_opt = app.add_option("--part-rec-height", part_rec_height, "Maximum rec section height");
        auto use_rec_part_opt = app.add_flag("--part-rec", use_rec_part, "Use rec partitioning");

        bool use_rec2_part = false;
        int part_rec2_max_depth = 0;
        double part_rec2_fluid_cost = 1.0;
        double part_rec2_solid_cost = 0.0;

        auto part_rec2_depth_opt = app.add_option("--part-rec2-max-depth", part_rec2_max_depth, "Maximum rec depth. Section count is 2^arg");
        auto use_rec2_part_opt = app.add_flag("--part-rec2", use_rec2_part, "Use rec partitioning");
        auto part_rec2_fluid_cost_opt = app.add_option("--part-rec2-fluid-cost", part_rec2_fluid_cost, "Cost of fluid point for partitioning. Default 1.0");
        auto part_rec2_solid_cost_opt = app.add_option("--part-rec2-solid-cost", part_rec2_solid_cost, "Cost of solid point for partitioning. Default 0.0");

        // Scheduling
        bool use_greedy = false;
        std::string greedy_cost_function = "Size";

        auto use_greedy_opt = app.add_flag("--sched-greedy", use_greedy, "Use greedy scheudling, default is round robin");
        auto greedy_cost_function_opt = app.add_option("--greedy-cost", greedy_cost_function, "The cost function to use. Supported is Size and Model.");
        
        double sched_greedy_fluid_cost = 1.0;
        double sched_greedy_solid_cost = 1.0;
        auto sched_greedy_fluid_cost_opt = app.add_option("--greedy-fluid-cost", sched_greedy_fluid_cost, "Cost of fluid point for greedy scheduling. Default 1.0");
        auto sched_greedy_solid_cost_opt = app.add_option("--greedy-solid-cost", sched_greedy_solid_cost, "Cost of solid point for greedy scheduling. Default 1.0");


        // CLI input file
        app.add_option("-f,--file", lattice_file, "Path to .pbm file with input structure")->check(CLI::ExistingFile);

        CLI11_PARSE(app, argc, argv);

        std::function<void(Section& section, int row, int col, int i, float& delta_N)> force_applier = No_force{};
        if (force_type == "channel") {
            force_applier = Channel_force(channel_force_lower, channel_force_upper);
        }

        std::function<std::vector<Section>(Lattice&)> sectioner = [](Lattice& l){return std::vector<Section>{Section(0,0, l.height, l.width)};};
        if (use_grid_part) {
            sectioner = [=](Lattice& lattice) {return partition_grid(lattice, {part_grid_height, part_grid_width});};
        }
        if (use_rec_part) {
            sectioner = [=](Lattice& lattice) {return partition_rec(lattice, {part_rec_height, part_rec_width});};
        }
        if (use_rec2_part) {
            sectioner = [=](Lattice& lattice) {return partition_rec2(lattice, part_rec2_max_depth, part_rec2_fluid_cost, part_rec2_solid_cost);};
        }
        if (!use_grid_part && !use_rec_part && !use_rec2_part) {
            std::cout << "No partitioning specified, only one full section will be used \n";
        }

#pragma endregion // CLI parser
        
#pragma region // Reserve space for time measurements

        // Setup timing 
        std::chrono::duration<long, std::nano> image_load_time{0};
        std::chrono::duration<long, std::nano> sectioning_time{0};
        std::chrono::duration<long, std::nano> exchange_detect_time{0};
        std::chrono::duration<long, std::nano> scheduling_time{0};

        // Timepoints used when timing loop
        std::chrono::steady_clock::time_point t_loop_begin;
        std::chrono::steady_clock::time_point t_collide_begin;
        std::chrono::steady_clock::time_point t_propagate_begin;
        std::chrono::steady_clock::time_point t_border_exch_begin;
        std::chrono::steady_clock::time_point t_sync_begin;

        // Loop timing
#ifdef TIME_COMP_LOOP
        std::vector<std::chrono::duration<long, std::nano>> comp_loop_times;
        comp_loop_times.reserve(MAX_ITER);
#endif //TIME_COMP_LOOP

#ifdef TIME_COLLIDE
        std::vector<std::chrono::duration<long, std::nano>> collide_times;
        collide_times.reserve(MAX_ITER);
#endif //TIME_COLLIDE

#ifdef TIME_BORDER_EXCHANGE
        std::vector<std::chrono::duration<long, std::nano>> exchange_times;
        exchange_times.reserve(MAX_ITER);
#endif //TIME_BORDER_EXCHANGE

#ifdef TIME_PROPAGATE
        std::vector<std::chrono::duration<long, std::nano>> progagate_time;
        progagate_time.reserve(MAX_ITER);
#endif //TIME_PROPAGATE

#ifdef TIME_SYNC_BEFORE_EXCHANGE
        std::vector<std::chrono::duration<long, std::nano>> sync_times;
        sync_time.reserve(MAX_ITER);
#endif //TIME_SYNC_BEFORE_EXCHANGE

#pragma endregion // Timing space

#pragma region // Initialization, setup and scheduling on master
        if (rank == 0) {
            auto t1 = std::chrono::steady_clock::now();
            global_lattice = load_from_image(lattice_file); // load_from_image("src/image.pbm");
            auto t2 = std::chrono::steady_clock::now();
            image_load_time = t2 - t1;

            global_section = Section{{0,0}, global_lattice.height, global_lattice.width};

            // Initialize it with an even distirbution
            initialize_even_distribution(global_lattice);

            // Partition
            auto t3 = std::chrono::steady_clock::now();
            
            all_sections = sectioner(global_lattice);

            auto t4 = std::chrono::steady_clock::now();
            sectioning_time = t4 - t3;

            // Determine required exchanges
            auto t5 = std::chrono::steady_clock::now();
            all_exchanges = determine_exchanges(all_sections, global_lattice);
            auto t6 = std::chrono::steady_clock::now();
            exchange_detect_time = t6 - t5;

            // Schedule the task onto ranks
            auto t7 = std::chrono::steady_clock::now();

            if (use_greedy) {
                if (greedy_cost_function == "Size") {
                    rank_assignment = schedule_greedy(all_sections, size, [&](const Section& s){return s.width * s.height;});
                } else 
                if (greedy_cost_function == "Model") {
                    rank_assignment = schedule_greedy(all_sections, size, [&](const Section& s){
                        // Iterate the lattice, and apply the weights:
                        double cost = 0;
                        for(int row = 0; row < s.height; row++) {
                            for(int col = 0; col < s.width; col++) {
                                switch (global_lattice.at(s.to_global_row(row), s.to_global_col(col)).type)
                                {
                                case Point_type::Fluid:
                                    cost += sched_greedy_fluid_cost;
                                    break;
                                case Point_type::Solid:
                                    cost += sched_greedy_solid_cost;
                                    break;
                                default:
                                    break;
                                }
                            }
                        }

                        return cost;
                    });
                } else {
                    std::cerr << "Invalid cost function for scheduler";
                    exit(1);
                }
            } else {
                rank_assignment = schedule_round_robin(all_sections, size);
            }


            auto t8 = std::chrono::steady_clock::now();
            scheduling_time = t8 - t7;

            // Map sections onto ranks, needed for exchange resolution
            section_assignment.resize(all_sections.size());
            for (int i = 0; i < rank_assignment.size(); i++) {
                for(int j = 0; j < rank_assignment[i].size(); j++) {
                    section_assignment[rank_assignment[i][j]] = i;
                }
            }

            //export_global_state(global_lattice, all_sections, all_exchanges, rank_assignment);
            //export_sections_as_tikz(global_lattice, all_sections, all_exchanges, rank_assignment, "tikz_export.tex");

            std::ofstream setup_timing_fs("./output/setup_timing.txt");
            setup_timing_fs << "\"load image time\"," << image_load_time.count() << "\n";
            setup_timing_fs << "\"section time\"," << sectioning_time.count() << "\n";
            setup_timing_fs << "\"exchange detect time\"," << exchange_detect_time.count() << "\n";
            setup_timing_fs << "\"schedule time\"," << scheduling_time.count() << "\n";
        }
#pragma endregion // Initialization, setup and scheduling

#pragma region // Print program info
        if (rank == 0) {
            // Print state of all flags
            // CLI args
            std::cout << "MAX_ITER" << ":" << MAX_ITER << "\n";
            std::cout << "channel_force_lower" << ":" << channel_force_lower << "\n";
            std::cout << "channel_force_upper" << ":" << channel_force_upper << "\n";
            std::cout << "force_type" << ":" << force_type << "\n";
            std::cout << "lattice_file" << ":" << lattice_file << "\n";

            #ifdef _OPENMP
            std::cout << "openmp_threads:" << omp_get_max_threads() << "\n";
            #endif // OPENMP
        }
#pragma endregion

#ifdef PREPROCESS_EXPORT
        if (rank == 0) {
            // We need to write out sections, and their exchanges, as well as the assignment. 
            // Write out sections
            std::ofstream section_output("./output/sections.dat");
            for (Section &sec : all_sections)
            {
                section_output << sec.tl_col << " " << sec.tl_row << " " << sec.br_col << " " << sec.br_row << '\n';
            }
            section_output.close();
            std::ofstream exchange_output("./output/exchanges.dat");
            for (int i = 0; i < all_exchanges.size(); i++)
            {
                for(auto ex: all_exchanges[i].send) {
                    exchange_output << i << " " << "S" << " " << ex.where << " " << ex.tag << " " << ex.section.tl_col << " " << ex.section.tl_row << " " << ex.section.br_col << " " << ex.section.br_row << '\n';
                }
                for(auto ex: all_exchanges[i].recv) {
                    exchange_output << i << " " << "R" << " " << ex.where << " " << ex.tag << " " << ex.section.tl_col << " " << ex.section.tl_row << " " << ex.section.br_col << " " << ex.section.br_row << '\n';
                }
            }
            exchange_output.close();
            
            std::ofstream schedule_output("./output/schedule.dat");
            for (int i = 0; i < section_assignment.size(); i++)
            {
                schedule_output << i << " " << section_assignment[i] << "\n";
            }
            schedule_output.close();
            
        }
#endif // PREPROCESS_EXPORT

#ifdef PREPROCESS_ONLY
        MPI_Finalize();
        exit(0);
#endif //PREPROCESS_ONLY

#pragma region // Scatter sections, exchange data etc
        if (rank == 0) {
            for (int i = 1; i < size; i++) {
                send_initial_scatter(i, rank_assignment, global_lattice, all_exchanges, all_sections, lat_mpi_type_fac);
                MPI_Barrier(MPI_COMM_WORLD);
            }
        }   
        else {
            for (int i = 1; i < size; i++) {
                if (i == rank) {
                    receive_initial_scatter(section_count, section_ids, sections, exchanges, lattices, lat_mpi_type_fac, rank, size);
                }
                MPI_Barrier(MPI_COMM_WORLD);
            }            
        }

        // Copy to self for the local lattices.
        if (rank == 0) {
            section_count = rank_assignment[0].size();
            sections.reserve(section_count);
            std::sort(rank_assignment[0].begin(), rank_assignment[0].end());
            
            section_ids = rank_assignment[0];

            for (int i: rank_assignment[0]){
                sections.push_back(all_sections[i]);
            }

            exchanges.reserve(section_count);
            for (int j = 0; j < section_count; j++) {
                exchanges.push_back(all_exchanges[rank_assignment[0][j]]);
            }

            // Here we copy the actual lattices
            lattices.reserve(section_count);
            for (int j = 0; j < section_count; j++) {
                const Section& cur_sec = sections[j];
                Lattice l = Lattice(cur_sec.height, cur_sec.width);
                for (int row = -1; row < cur_sec.height + 1; row++) {
                    for (int col = -1; col < cur_sec.width + 1; col++) {
                        l.at(row, col) = global_lattice.at(cur_sec.to_global_row(row), cur_sec.to_global_col(col));
                    }
                }
                lattices.push_back(std::move(l)); 
            }
            std::cout << "Rank 0 copied to self" << std::endl;
        }

        // Bcast of assignments, used for exchange resolution
        if (rank == 0) {
            int sec_assign_c = section_assignment.size();
            MPI_Bcast(&sec_assign_c, 1, MPI_INT, 0, MPI_COMM_WORLD);
            MPI_Bcast(section_assignment.data(), sec_assign_c, MPI_INT, 0, MPI_COMM_WORLD);
        } else {
            int sec_assign_c;
            MPI_Bcast(&sec_assign_c, 1, MPI_INT, 0, MPI_COMM_WORLD);
            section_assignment.resize(sec_assign_c);
            MPI_Bcast(section_assignment.data(), sec_assign_c, MPI_INT, 0, MPI_COMM_WORLD);
        }

        // Bcast global section data, we need it for size
        MPI_Bcast(&global_section, 1, section_type, 0, MPI_COMM_WORLD);
#pragma endregion // Scatter data to all ranks

#pragma region // Main compute

        std::cout << "Rank " << rank << " reached main compute loop" << std::endl;

#ifndef DISABLE_BORDER_EXCH
        border_exchange(sections, lattices, exchanges, section_assignment, lat_mpi_type_fac, rank);
#endif //DISABLE_BORDER_EXCH



        MPI_Barrier(MPI_COMM_WORLD);

#ifdef TIME_COMP
        std::chrono::steady_clock::time_point t_comp_begin = std::chrono::steady_clock::now();
#endif //TIME_COMP

        for(int iter = 0; iter < MAX_ITER; iter++) {
#ifdef PRINT_INTERMEDIATE_LATTICE
            if (iter % 10 == 0) {
                gather_sections(rank_assignment, all_sections, global_lattice, section_ids, sections, lattices, section_assignment, lat_mpi_type_fac, rank, size);

                if (rank == 0) {
                    std::stringstream s;
                    s << "./output/latice_iter_" << iter << ".txt";
                    std::ofstream final_lat_ofs{s.str()};
                    write_lattice_to_stream(final_lat_ofs, global_lattice);
                    final_lat_ofs.close();
                }
            }
#endif //PRINT_INTERMEDIATE_LATTICE

#ifdef TIME_COMP_LOOP
            t_loop_begin = std::chrono::steady_clock::now(); 
#endif //TIME_COMP_LOOP

            // Collide
#ifdef TIME_COLLIDE
            t_collide_begin = std::chrono::steady_clock::now();
#endif //TIME_COLLIDE
            
            //TODO: Use OpenMP task?
#ifdef DATA_P
            #pragma omp parallel for
#endif //DATA_P
#ifdef TASKP
        #pragma omp parallel
        #pragma omp single nowait
        {
#endif // TASKP
        for(int i = 0; i < section_count; i++){
            collide2(Raster(&sections[i], &lattices[i]), force_applier);
        }
#ifdef TASKP
        }
#endif // TASKP


#ifdef TIME_COLLIDE
            collide_times.push_back(std::chrono::steady_clock::now() - t_collide_begin);
#endif //TIME_COLLIDE



#ifndef DISABLE_BORDER_EXCH

#ifdef SYNC_BEFORE_EXCHANGE
#ifdef TIME_SYNC_BEFORE_EXCHANGE
            // Start time
            t_sync_begin = std::chrono::steady_clock::now();
#endif //TIME_SYNC_BEFORE_EXCHANGE

            MPI_Barrier(MPI_COMM_WORLD);

#ifdef TIME_SYNC_BEFORE_EXCHANGE
            sync_times.push_back(std::chrono::steady_clock::now() - t_sync_begin);
#endif //TIME_SYNC_BEFORE_EXCHANGE

#endif //SYNC_BEFORE_EXCHANGE

#ifdef TIME_BORDER_EXCHANGE
            t_border_exch_begin = std::chrono::steady_clock::now();
#endif //TIME_BORDER_EXCHANGE

            border_exchange(sections, lattices, exchanges, section_assignment, lat_mpi_type_fac, rank);

#ifdef TIME_BORDER_EXCHANGE
            exchange_times.push_back(std::chrono::steady_clock::now() - t_border_exch_begin);
#endif //TIME_BORDER_EXCHANGE

#endif //DISABLE_BORDER_EXCH
            
            // Propagate 
#ifdef TIME_PROPAGATE
            t_propagate_begin = std::chrono::steady_clock::now();
#endif //TIME_PROPAGATE

            //TODO: Use OpenMP task?
#ifdef DATA_P
        #pragma omp parallel for
#endif //DATA_P
#ifdef TASKP
        #pragma omp parallel
        #pragma omp single nowait
        {
#endif // TASKP
        for(int i = 0; i < section_count; i++){
            propagate(Raster(&sections[i], &lattices[i]), global_section);
        }
#ifdef TASKP
        }
#endif // TASKP

#ifdef TIME_PROPAGATE
            progagate_time.push_back(std::chrono::steady_clock::now() - t_propagate_begin);
#endif //TIME_PROPAGATE

#ifdef TIME_COMP_LOOP
            comp_loop_times.push_back(std::chrono::steady_clock::now() - t_loop_begin);
#endif //TIME_COMP_LOOP 

        }
#pragma endregion // Main compute

#pragma region // Export result

        #ifdef TIME_COMP
        std::chrono::duration<long, std::nano> t_comp(std::chrono::steady_clock::now() - t_comp_begin);
        { 
            // Write out the times
            std::stringstream file_name;
            file_name << "./output/compute_time_" << rank << ".txt";
            std::string file_name_str = file_name.str();
            std::cout << "Writing compute time to " << file_name_str << "\n";
            std::ofstream comp_time_file(file_name_str);
            comp_time_file << t_comp.count();
        }
        #endif //TIME_COMP

        std::cout << "Compute complete on " << rank << std::endl;

#ifdef TIME_COMP_LOOP
        {
            // Write out the times
            std::stringstream file_name;
            file_name << "./output/loop_timing_" << rank << ".txt";
            std::string file_name_str = file_name.str();
            std::cout << "Writing collide times to " << file_name_str << "\n";
            std::ofstream loop_time_file(file_name_str);
            for (auto t: comp_loop_times) {
                loop_time_file << t.count() << '\n';
            }
        }
#endif //TIME_COMP_LOOP

#ifdef TIME_PROPAGATE
        {
            // Write out the times
            std::stringstream file_name;
            file_name << "./output/propagate_timing_" << rank << ".txt";
            std::ofstream loop_time_file(file_name.str());
            for (auto&& t: progagate_time) {
                loop_time_file << t.count() << '\n';
            }
        }
#endif //TIME_PROPAGATE
            
#ifdef TIME_COLLIDE
        {
            // Write out the times
            std::stringstream file_name;
            file_name << "./output/collide_timing_" << rank << ".txt";
            std::string file_name_str = file_name.str();
            std::cout << "Writing collide times to " << file_name_str << "\n";
            std::ofstream loop_time_file(file_name_str);
            for (auto&& t: collide_times) {
                loop_time_file << t.count() << '\n';
            }
        }
#endif //TIME_COLLIDE

#ifdef TIME_BORDER_EXCHANGE
        {
            // Write out the times
            std::stringstream file_name;
            file_name << "./output/exchange_timing_" << rank << ".txt";
            std::ofstream loop_time_file(file_name.str());
            for (auto&& t: exchange_times) {
                loop_time_file << t.count() << '\n';
            }
        }
#endif //TIME_BORDER_EXCHANGE

#ifdef TIME_SYNC_BEFORE_EXCHANGE
        {
            // Write out the times
            std::stringstream file_name;
            file_name << "./output/sync_timing_" << rank << ".txt";
            std::ofstream sync_time_file(file_name.str());
            for (auto&& t: sync_times) {
                sync_time_file << t.count() << '\n';
            }
        }
#endif //TIME_SYNC_BEFORE_EXCHANGE

#ifdef PRINT_FINAL_LATTICE
        gather_sections(rank_assignment, all_sections, global_lattice, section_ids, sections, lattices, section_assignment, lat_mpi_type_fac, rank, size);
        if (rank == 0) {
            std::ofstream final_lat_ofs{"./output/final_lat.txt"};
            write_lattice_to_stream(final_lat_ofs, global_lattice);
            final_lat_ofs.close();
        }
#endif //PRINT_FINAL_LATTICE
#pragma endregion // Export result

        free_mpi_types();
    }
    MPI_Finalize();

    std::cout << "Finalize complete on rank " << rank << std::endl;

    return res + 0;
}
