
#include "mpi_exchange.hpp"

MPI_Datatype section_type;
MPI_Datatype exchange_type;
MPI_Datatype lpoint_type;

int init_mpi_types() {
    section_type = init_section_type();
    MPI_Type_commit(&section_type);
    exchange_type = init_exchange_type();
    MPI_Type_commit(&exchange_type);
    lpoint_type = init_point_type();
    MPI_Type_commit(&lpoint_type);
    return 0;
}

int free_mpi_types() {
    MPI_Type_free(&section_type);
    MPI_Type_free(&exchange_type);
    MPI_Type_free(&lpoint_type);
    return 0;
}

void send_exchange_struct(const Section_exchange& exch, int rank, int tag) {
    // First send the sizes of the two arrays
    int sizes[2] = {(int) exch.recv.size(), (int) exch.send.size()};
    MPI_Send(&sizes, 2, MPI_INT, rank, 1000 | tag, MPI_COMM_WORLD);

    // Then we send the content of the two vectors
    MPI_Send(exch.recv.data(), exch.recv.size(), exchange_type, rank, 2000 | tag, MPI_COMM_WORLD);
    MPI_Send(exch.send.data(), exch.send.size(), exchange_type, rank, 3000 | tag, MPI_COMM_WORLD);
}

Section_exchange recv_exchange_struct(int rank, int tag) {
    Section_exchange exch;
    int sizes[2];
    MPI_Recv(&sizes, 2, MPI_INT, rank, 1000 | tag, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    exch.recv.resize(sizes[0]);
    exch.send.resize(sizes[1]);
    MPI_Recv(exch.recv.data(), sizes[0], exchange_type, rank, 2000 | tag, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    MPI_Recv(exch.send.data(), sizes[1], exchange_type, rank, 3000 | tag, MPI_COMM_WORLD, MPI_STATUS_IGNORE);

    return exch;
}

void border_exchange(   std::vector<Section>& sections, 
                        std::vector<Lattice> &lattices, 
                        std::vector<Section_exchange> &exchanges,
                        std::vector<int> &section_assignment,
                        Lattice_type_factory &lat_mpi_type_fac,
                        int rank) {
    // Border exchange
    std::vector<MPI_Request*> send_reqs;
    std::vector<MPI_Request*> receive_reqs;
    for(int i = 0; i < sections.size(); i++) {
        const auto& curr_exch = exchanges[i];
        auto& lat = lattices[i];
        const auto& sec = sections[i];

        // Receive
        for (int j = 0; j < curr_exch.recv.size(); j++){ 
            const auto& exch = curr_exch.recv[j];
            const auto& exch_sec = exch.section;
            const int src = section_assignment[exch.where];

            MPI_Request *req = new MPI_Request;
            MPI_Irecv(&lat.at(exch_sec.tl_row, exch_sec.tl_col), 1, lat_mpi_type_fac.subsec_type(lat, exch_sec, 0), src, exch.tag, MPI_COMM_WORLD, req);
            receive_reqs.push_back(req);
        }

        // Send
        for (int j = 0; j < curr_exch.send.size(); j++){ 
            const auto& exch = curr_exch.send[j];
            const auto& exch_sec = exch.section;
            const int dest = section_assignment[exch.where];

            MPI_Request *req = new MPI_Request();
            MPI_Isend(&lat.at(exch_sec.tl_row, exch_sec.tl_col), 1, lat_mpi_type_fac.subsec_type(lat, exch_sec, 0), dest, exch.tag, MPI_COMM_WORLD, req);
            send_reqs.push_back(req);
        }
    }
    
    // Wait for send to complete
    for (int j = 0; j < send_reqs.size(); j++) {
        MPI_Wait(send_reqs[j], MPI_STATUS_IGNORE);
        delete send_reqs[j];
    }
    // Wait for all receives to complete
    for (int j = 0; j < receive_reqs.size(); j++) {
        MPI_Wait(receive_reqs[j], MPI_STATUS_IGNORE);
        delete receive_reqs[j];
    }
}

void gather_sections(   std::vector<std::vector<int>>& rank_assignment,
                        std::vector<Section>& all_sections,
                        Lattice& global_lattice,
                        std::vector<int>& section_ids,
                        std::vector<Section>& sections, 
                        std::vector<Lattice> &lattices,
                        std::vector<int> &section_assignment,
                        Lattice_type_factory &lat_mpi_type_fac,
                        int rank, int size) {
    // Gather the result
    if (rank == 0) {
        for (int i = 1; i < size; i++) {
            // Then we receive the data for the section aka actual lattice
            for (int j = 0; j < rank_assignment[i].size(); j++) {
                const Section& cur_sec =  all_sections[rank_assignment[i][j]];
                //std::cout << "Receiving data for section " << rank_assignment[i][j] << " from rank " << i << std::endl;

                MPI_Recv(&(global_lattice.at(cur_sec.tl_row, cur_sec.tl_col)), 1, lat_mpi_type_fac.subsec_type(global_lattice, cur_sec, 0), i, j, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            }

            MPI_Barrier(MPI_COMM_WORLD);
        }
    }   
    else {
        for (int i = 1; i < size; i++) {
            if (i == rank) {
                // Here we should send back the actual lattices
                for (int j = 0; j < sections.size(); j++) {
                    //std::cout << "Sending data for section " << section_ids[j] << " from rank " << rank << std::endl;
                    MPI_Send(&lattices[j].at(0, 0), 1, lat_mpi_type_fac.subsec_type(lattices[j], sections[j], 0), 0, j, MPI_COMM_WORLD);
                }
            }
            MPI_Barrier(MPI_COMM_WORLD);
        }
    }

    // Rank 0 copy to self
    if (rank == 0) {
        for (int i = 0; i < sections.size(); i++){ 
            const Section& cur_sec =  sections[i];
            for (int row = 0; row < cur_sec.height; row++) {
                for (int col = 0; col < cur_sec.width; col++) {
                    global_lattice.at(cur_sec.to_global_row(row), cur_sec.to_global_col(col)) = lattices[i].at(row, col);
                }
            }
        }

    }
}

void receive_initial_scatter(int &section_count, std::vector<int> &section_ids, std::vector<Section> &sections, std::vector<Section_exchange> &exchanges, std::vector<Lattice> &lattices, Lattice_type_factory &lat_mpi_type_fac,
                        int rank, int size) {
    // First we receive the number of sections we are to get
    MPI_Recv(&section_count, 1, MPI_INT, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    std::cout << "Received " << section_count << " as section count on " << rank << std::endl;

    // We then get the section ids
    section_ids.resize(section_count);
    MPI_Recv(section_ids.data(), section_count, MPI_INT, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    
    // We get the 
    sections.resize(section_count);
    MPI_Recv(sections.data(), section_count, section_type, 0, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE); 
    std::cout << "Received assigned sections on rank " << rank << std::endl;

    // Receive the exchange info
    exchanges.reserve(section_count);
    for(int i = 0; i < section_count; i++) {
        exchanges.push_back(recv_exchange_struct(0, i));
        std::cout << "Received exchange struct " << i << " on rank " << rank << std::endl;
    }

    // Here we should receve the actual lattices
    lattices.reserve(section_count);
    for (int j = 0; j < section_count; j++) {
        Lattice l(sections[j].height, sections[j].width);
        MPI_Recv(&l.atd(0, 0), 1, lat_mpi_type_fac.subsec_type(l, sections[j], 1), 0, j, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        std::cout << "Received data for section " << j << " on rank " << rank << std::endl;
        lattices.push_back(std::move(l));
    }
}

void send_initial_scatter(int i, std::vector<std::vector<int>>& rank_assignment, Lattice& global_lattice, std::vector<Section_exchange>& all_exchanges, std::vector<Section>& all_sections, Lattice_type_factory &lat_mpi_type_fac)
{ 
    // First we send number of sections, so space to the info can be set aside
    int assigned_section_c = rank_assignment[i].size();
    std::cout << "Sending " << assigned_section_c << " as section count to " << i << std::endl;
    MPI_Send(&assigned_section_c, 1, MPI_INT, i, 0, MPI_COMM_WORLD);

    // Sort the rank assignment. This might be wasted.
    std::sort(rank_assignment[i].begin(), rank_assignment[i].end());
    
    // We organize the sections into a contigous vector
    std::vector<Section> assigned_sections;
    assigned_sections.reserve(assigned_section_c);
    for (int i: rank_assignment[i]){
        assigned_sections.push_back(all_sections[i]);
    }

    // We send the rank assignment, needed for quick lookup of local border exchange
    MPI_Send(rank_assignment[i].data(), assigned_section_c, MPI_INT, i, 0, MPI_COMM_WORLD);

    // Then the actual section data
    std::cout << "Sending assigned sections to rank " << i << std::endl;
    MPI_Send(assigned_sections.data(), assigned_section_c, section_type, i, 0, MPI_COMM_WORLD);
    
    // For each section:
    // Then we send all exchange info
    for (int j = 0; j < assigned_section_c; j++) {
        std::cout << "Sending exchange struct " << rank_assignment[i][j] << " to rank " << i << std::endl;
        send_exchange_struct(all_exchanges[rank_assignment[i][j]], i, j);
    }

    // Then we send the data for the section aka actual lattice
    for (int j = 0; j < assigned_section_c; j++) {
        const Section& cur_sec = assigned_sections[j];
        std::cout << "Sending data for section " << rank_assignment[i][j] << " to rank " << i << std::endl;

        MPI_Send(&(global_lattice.at(cur_sec.tl_row - 1, cur_sec.tl_col - 1)), 1, lat_mpi_type_fac.subsec_type(global_lattice, cur_sec , 1), i, j, MPI_COMM_WORLD);
    }
}