#include "part_rec2.hpp"

#include "doctest.hpp"

std::vector<Section> split_vertical(Lattice& l, Section s, double fluid_cost, double solid_cost) {
    // Loop over all values and sum it up. If all fluid points lie on the same vertical, split in the middle

    // We'll do vertical split, so sum columns
    double sum = 0;
    std::vector<double> column_sums;
    for(int col = 0; col < s.width; col++) {
        double col_sum = 0;
        for (int row = 0; row < s.height; row++) {
            if (l.at(s.to_global_row(row), s.to_global_col(col)).type == Point_type::Fluid) {
                col_sum += fluid_cost;
            } else if (l.at(s.to_global_row(row), s.to_global_col(col)).type == Point_type::Solid) {
                col_sum += solid_cost;
            }
        }
        sum += col_sum;
        column_sums.push_back(col_sum);
    }

    // We'll see if all values are in the same column. 
    if (std::count_if(std::begin(column_sums), std::end(column_sums), [](double v){return v != 0;}) < 2) {
        // Single_column is true, a cut in the middle is as good as any
        // Cut middle
        // Return
        int width_middle = static_cast<int>(ceil(s.width / 2.0));
        return std::vector<Section>{
            Section({0,0}, {s.height-1, width_middle-1}),
            Section({0,width_middle}, {s.height-1, s.width-1})
        };
    }

    // Find the middle of the column vector. 
    double middle = sum / 2.0;

    int split_idx = 0;
    double current_count = 0;
    while (current_count < middle) {
        current_count += column_sums[split_idx];
        split_idx++;
    }

    // We have the column to split on in split_idx
    return std::vector<Section>{
        Section({s.tl_row, s.tl_col}, {s.br_row, s.tl_col + split_idx-1}),
        Section({s.tl_row, s.tl_col + split_idx}, {s.br_row, s.br_col})
    };
}

std::vector<Section> split_horisontal(Lattice&l, Section s, double fluid_cost, double solid_cost) {
    // Loop over all values and sum it up. If all fluid points lie on the same horizontal, split in the middle
    double sum = 0;
    std::vector<double> row_sums;
    for (int row = 0; row < s.height; row++) {
        double row_sum = 0;
        for(int col = 0; col < s.width; col++) {
            if (l.at(s.to_global_row(row), s.to_global_col(col)).type == Point_type::Fluid) {
                row_sum += fluid_cost;
            } else if (l.at(s.to_global_row(row), s.to_global_col(col)).type == Point_type::Solid) {
                row_sum += solid_cost;
            }
        }
        sum += row_sum;
        row_sums.push_back(row_sum);
    }

    // We'll see if all values are in the same column. 
    if (std::count_if(std::begin(row_sums), std::end(row_sums), [](double v){return v != 0;}) < 2) {
        // Single_column is true, a cut in the middle is as good as any
        // Cut middle
        // Return
        int height_middle = static_cast<int>(ceil(s.height / 2.0));
        return std::vector<Section>{
            Section({0,0}, {height_middle - 1, s.width-1}),
            Section({height_middle, 0}, {s.height-1, s.width-1})
        };
    }

    // Find the middle of the column vector. 
    double middle = sum / 2.0;

    int split_idx = 0;
    double current_count = 0;
    while (current_count < middle) {
        current_count += row_sums[split_idx];
        split_idx++;
    }

    // We have the column to split on in split_idx
    return std::vector<Section>{
        Section({s.tl_row,s.tl_col}, {s.tl_row + split_idx-1, s.br_col}),
        Section({s.tl_row + split_idx, s.tl_col}, {s.br_row, s.br_col})
    };
}

std::vector<Section> split(Lattice& l, Section s, int depth, double fluid_cost, double solid_cost) {
    // Determine directions, depth % 2: 0 - Horizontal 1 - Vertical
    int scan_direction = depth % 2;

    return scan_direction == 0 ? split_horisontal(l, s, fluid_cost, solid_cost) : split_vertical(l, s, fluid_cost, solid_cost);
}

std::vector<Section> part_rec(Lattice& l, Section s, int depth, int max_depth, double fluid_cost, double solid_cost)
{
    // Stop condition
    if (depth > max_depth) {
        return std::vector<Section>{s};
    } 

    // Split and recurse
    auto sections = split(l, s, depth, fluid_cost, solid_cost);

    std::vector<Section> result;
    for (auto section : sections)
    {
        auto r = part_rec(l, section, depth+1, max_depth, fluid_cost, solid_cost);
        result.insert(std::end(result), std::begin(r), std::end(r));
    }

    return result;
}

std::vector<Section> partition_rec2(Lattice& lattice, int max_depth, double fluid_cost, double solid_cost) {
    // We start my making a section for the entire lattice
    Section root(0, 0, lattice.height, lattice.width);

    // Now, we should do recursion
    //We'll carry the depth, and use that do determine direction
    return part_rec(lattice, root, 1, max_depth, fluid_cost, solid_cost);
}

TEST_CASE("RecursiveSplitterReturnsSingleWithDepth0")
{
    Lattice l{8, 8};

    std::vector<Section> sections = partition_rec2(l, 0);

    CHECK_EQ(sections.size(), 1);
}

TEST_CASE("RecursiveSplitterReturnsTwoWithDepth1")
{
    Lattice l{8, 8};

    std::vector<Section> sections = partition_rec2(l, 1);

    CHECK_EQ(sections.size(), 2);
}

TEST_CASE("RecursiveSplitterReturnsFourWithDepth2")
{
    Lattice l{8, 8};

    std::vector<Section> sections = partition_rec2(l, 2);

    CHECK_EQ(sections.size(), 4);
}

TEST_CASE("RecursiveSplitterReturnsCorrectForTwoSection")
{
    Lattice l{8, 8};

    std::vector<Section> sections = partition_rec2(l, 1);

    // Expected sections
    Section first(0, 0, 8, 4);
    Section second(0, 4, 8, 4);
    
    CHECK_EQ(sections[0].tl_row, first.tl_row);
    CHECK_EQ(sections[0].tl_col, first.tl_col);
    CHECK_EQ(sections[0].br_row, first.br_row);
    CHECK_EQ(sections[0].br_col, first.br_col);

    CHECK_EQ(sections[1].tl_row, second.tl_row);
    CHECK_EQ(sections[1].tl_col, second.tl_col);
    CHECK_EQ(sections[1].br_row, second.br_row);
    CHECK_EQ(sections[1].br_col, second.br_col);
}

TEST_CASE("RecursiveSplitterReturnsCorrectForFourSection")
{
    Lattice l{8, 8};

    std::vector<Section> sections = partition_rec2(l, 2);

    // Expected sections
    Section first(0, 0, 4, 4);
    Section second(4, 0, 4, 4);
    Section third(0, 4, 4, 4);
    Section fourth(4, 4, 4, 4);
    
    CHECK_EQ(sections[0].tl_row, first.tl_row);
    CHECK_EQ(sections[0].tl_col, first.tl_col);
    CHECK_EQ(sections[0].br_row, first.br_row);
    CHECK_EQ(sections[0].br_col, first.br_col);

    CHECK_EQ(sections[1].tl_row, second.tl_row);
    CHECK_EQ(sections[1].tl_col, second.tl_col);
    CHECK_EQ(sections[1].br_row, second.br_row);
    CHECK_EQ(sections[1].br_col, second.br_col);

    CHECK_EQ(sections[2].tl_row, third.tl_row);
    CHECK_EQ(sections[2].tl_col, third.tl_col);
    CHECK_EQ(sections[2].br_row, third.br_row);
    CHECK_EQ(sections[2].br_col, third.br_col);

    CHECK_EQ(sections[3].tl_row, fourth.tl_row);
    CHECK_EQ(sections[3].tl_col, fourth.tl_col);
    CHECK_EQ(sections[3].br_row, fourth.br_row);
    CHECK_EQ(sections[3].br_col, fourth.br_col);
}


TEST_CASE("RecursiveSplitterReturnsCorrectForTwoSectionNonDefaultWeights")
{
    Lattice l{8, 8};
    for(int row = 0; row < 8; row++){
        for(int col = 0; col < 4; col++) {
            l.at(row, col).type = Point_type::Solid;
        }
    }
    CHECK_EQ(l.at(0, 0).type, Point_type::Solid);

    std::vector<Section> sections = partition_rec2(l, 1, 1.0, 0.5);

    CHECK_EQ(sections.size(), 2);

    // Expected sections
    Section first(0, 0, 8, 5);
    Section second(0, 5, 8, 3);
    
    CHECK_EQ(sections[0].tl_row, first.tl_row);
    CHECK_EQ(sections[0].tl_col, first.tl_col);
    CHECK_EQ(sections[0].br_row, first.br_row);
    CHECK_EQ(sections[0].br_col, first.br_col);
}