#include <mpi.h>
#include <cstdlib>
#include <cstring>

#include <iostream>
#include <fstream>
#include <sstream>

struct Config
{
    int start;
    int stop;
    int steps;
    double W;
    int N;
    std::string base_filename;
    int warmup;
};

struct TestRunData
{
    int initiator;
    int mirror;
};

int 
message_size(int start, int stop, int steps, int step)
{
    return start + ((stop - start) / (steps - 1)) * step;
}

void 
busy_wait(double duration)
{
    double start = MPI_Wtime();
    while((MPI_Wtime() - start) < duration){}
}

double 
initiator_execute_roundtrip(Config conf, int step, double W, std::ostream &output_stream, TestRunData trd)
{
    // Prepare what we need
    int packet_size = message_size(conf.start, conf.stop, conf.steps, step);
    // std::cout << "Sending packetsize " << packet_size << std::endl;
    // Send the packetsize
    MPI_Send(&packet_size, 1, MPI_INT, trd.mirror, 0, MPI_COMM_WORLD);

    char *data = new char[packet_size];

    // std::cout << "Doing the busy wait" << std::endl;
    // We want to make sure we never wait for mirror, so delay the initiation. P1 should be ahead anyways

    for (int j = 0; j < conf.warmup; j++)
    {
        MPI_Send(data, packet_size, MPI_CHAR, trd.mirror, 0, MPI_COMM_WORLD);
        MPI_Recv(data, packet_size, MPI_CHAR, trd.mirror, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
    }

    busy_wait(0.1);

    // Start timer
    // std::cout << "Starting timer" << std::endl;
    double time_accumulator = 0;

    double start_time = MPI_Wtime();
    for (int j = 0; j < conf.N; j++)
    {
        
        MPI_Send(data, packet_size, MPI_CHAR, trd.mirror, 0, MPI_COMM_WORLD);

        busy_wait(W);
        
        MPI_Recv(data, packet_size, MPI_CHAR, trd.mirror, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        
    }
    double end_time = MPI_Wtime();

    double average = (end_time - start_time) / conf.N;

    output_stream << std::scientific << step << "\t" << packet_size << "\t" << conf.W << "\t" << conf.N << "\t" << average << std::endl;
    
    delete[] data;
    
    return average;
}

void
initiator(Config conf, TestRunData trd)
{
    // We will perform steps *2  iterations of pingpong and record the time
    // One run per iteration will have w = 0, and one w = W
    // Each run will use a message size determined by start + (stop-start)/steps

    // Open files for the result
    std::stringstream stream_base_name;
    stream_base_name << conf.base_filename << "_" << trd.initiator << "_" << trd.mirror;
    std::ofstream w0_stream((stream_base_name.str() + "_w0.tsv"));
    std::ofstream w_stream((stream_base_name.str() + "_w.tsv"));

    double* results = new double[conf.steps * 2];

    for (int i = 0; i < conf.steps; i++)
    {
        results[2 * i] = initiator_execute_roundtrip(conf, i, 0, w0_stream, trd);
        results[2 * i + 1] = initiator_execute_roundtrip(conf, i, conf.W, w_stream, trd);
    }

    w0_stream.close();
    w_stream.close();

    delete[] results;
    return;
}

void
mirror(Config conf, TestRunData trd)
{
    for (int i = 0; i < conf.steps * 2; i++){
        // Prepare what we need
        int packet_size;
        MPI_Recv(&packet_size, 1, MPI_INT, trd.initiator, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
        char *data = new char[packet_size];

        for (int j = 0; j < conf.warmup; j++)
        {
            MPI_Recv(data, packet_size, MPI_CHAR, trd.initiator, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            MPI_Send(data, packet_size, MPI_CHAR, trd.initiator, 0, MPI_COMM_WORLD);
        }

        for (int j = 0; j < conf.N; j++){
            MPI_Recv(data, packet_size, MPI_CHAR, trd.initiator, 0, MPI_COMM_WORLD, MPI_STATUS_IGNORE);
            MPI_Send(data, packet_size, MPI_CHAR, trd.initiator, 0, MPI_COMM_WORLD);
        }

        delete[] data;
    }

    return;
}

void 
perform_bench_all(Config conf)
{
    int rank, size;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &size);

    for(int initRank = 0; initRank < size - 1; initRank++) 
    {
        for (int mirrorRank = initRank + 1; mirrorRank < size; mirrorRank++) {
            TestRunData trd = {
                .initiator = initRank,
                .mirror = mirrorRank
            };

            if (rank == initRank) {
                initiator(conf, trd);
            }
            if (rank == mirrorRank) {
                mirror(conf, trd);
            }
            MPI_Barrier(MPI_COMM_WORLD);
        }
    }
}

int 
find_string(int argc, char **argv, std::string compare)
{
    for (int i = 0; i < argc; i++) {
        if (strcmp(argv[i], compare.c_str()) == 0) {
            return i;
        }
    }
    return -1;
}

void
usage()
{
    std::cout << "Usage: loggps [options]" << std::endl;
    std::cout << "Options: " << std::endl;
    std::cout << "\t-h --help \t\tDisplay this usage info" << std::endl;
    std::cout << "\t--start value \tThe number of bytes to start with" << std::endl;
    std::cout << "\t--stop value \tThe number of bytes to end with" << std::endl;
    std::cout << "\t--step value \tThe number of steps to use" << std::endl;
    std::cout << "\t-W value \t\tThe W (busy wait) for initiator" << std::endl;
    std::cout << "\t-N value \t\tHow many iterations to average per step" << std::endl;
    std::cout << "\t-F string \t\tThe step for the output filenames" << std::endl;
    std::cout << std::endl;
}

int
main(int argc, char** argv)
{
    MPI_Init(&argc, &argv);

    int rank;
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    Config conf;
    int arg_index;

    arg_index = find_string(argc, argv, "-h") + find_string(argc, argv, "--help");
    if (argc % 2 == 0 || arg_index > -2) {
        if (rank == 0)
        {
            if (arg_index == -2)
            {
                std::cout << "Invalid number of params" << std::endl;
            }
            usage();
        }
        MPI_Finalize();
        exit(1);
    }

    // Look for start argument
    arg_index = find_string(argc, argv, "--start");
    conf.start = (arg_index > -1) ?  atoi(argv[arg_index + 1]) : 1;

    // Look for end argument
    arg_index = find_string(argc, argv, "--stop");
    conf.stop = (arg_index > -1) ? atoi(argv[arg_index + 1]) : 65000;

    // Look for step argument
    arg_index = find_string(argc, argv, "--steps");
    conf.steps = (arg_index > -1) ? atoi(argv[arg_index + 1]) : 100;

    // Look for W argument
    arg_index = find_string(argc, argv, "-W");
    conf.W = (arg_index > -1) ? atof(argv[arg_index + 1]) : 5*10e-4;

    // Look for N argument
    arg_index = find_string(argc, argv, "-N");
    conf.N = (arg_index > -1) ? atof(argv[arg_index + 1]) : 10;

    // Look for base filename argument
    arg_index = find_string(argc, argv, "-F");
    conf.base_filename = (arg_index > -1) ? std::string(argv[arg_index + 1]) : "data";

    // Look for warmup argument
    arg_index = find_string(argc, argv, "--warmup");
    conf.warmup = (arg_index > -1) ? atof(argv[arg_index + 1]) : 0;

    if(rank ==0) {
        std::cout << "Running LogGPS Benchmark" << std::endl;
        std::cout << "START: " << conf.start << " STOP: " << conf.stop << " STEPS: " << conf.steps << " W: " << conf.W << " N: " << conf.N << " WARMUP: " <<  conf.warmup << std::endl; 
    }
        
    
    double time_start = MPI_Wtime();
    perform_bench_all(conf);
    double total_duration = MPI_Wtime() - time_start;

    if (rank ==0)
    {
        std::cout << "Completed after " << total_duration << "s" <<std::endl;
    }

    MPI_Finalize();
    return 0;
}